# -*- coding: utf-8 -*-
# Derivcheck is robust and very sensitive tester for analytic derivatives.
# Copyright (C) 2017 Toon Verstraelen <Toon.Verstraelen@UGent.be>.
#
# This file is part of Derivcheck.
#
# Derivcheck is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Derivcheck is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""Unit tests for derivcheck."""

from nose.tools import assert_raises
import numpy as np

from derivcheck import diff_ridders, assert_deriv
from .basic_example import main as example_main


def test_ridders_corner_cases():
    with assert_raises(ValueError):
        diff_ridders(np.sin, 0.0, 0.0)
    with assert_raises(ValueError):
        diff_ridders(np.sin, 0.0, 1.0, con=0.9)
    with assert_raises(ValueError):
        diff_ridders(np.sin, 0.0, 1.0, safe=0.9)
    assert diff_ridders(np.sin, 0.0, 1.0, maxiter=0) == (None, None)


def test_ridders_simple():
    for arg in np.linspace(-1.0, 1.0, 15):
        deriv, error = diff_ridders(np.exp, arg, 0.1)
        assert error < 1e-10
        np.testing.assert_allclose(deriv, np.exp(arg))
        deriv, error = diff_ridders(np.sin, arg, 0.1)
        assert error < 1e-10
        np.testing.assert_allclose(deriv, np.cos(arg))


def _check_assert_deriv_0d_harm(nrep, arg_shape, numtested):
    _function = lambda arg: 0.5 * np.sum(arg**2)
    _gradient = lambda arg: arg
    for _ in range(nrep):
        origin = np.random.normal(0, 1, arg_shape)
        assert assert_deriv(_function, _gradient, origin) == numtested


def test_assert_deriv_0d_harm():
    yield _check_assert_deriv_0d_harm, 10, (), 1
    yield _check_assert_deriv_0d_harm, 10, (10, ), 10
    yield _check_assert_deriv_0d_harm, 10, (3, 4), 12
    yield _check_assert_deriv_0d_harm, 10, (2, 3, 4), 24


def _check_assert_deriv_0d_exp(nrep, arg_shape, numtested):
    _function = lambda arg: np.exp(arg).sum()
    for _ in range(nrep):
        origin = np.random.uniform(-1.0, 1.0, arg_shape)
        assert assert_deriv(_function, np.exp, origin) == numtested


def test_assert_deriv_0d_exp():
    yield _check_assert_deriv_0d_exp, 10, (), 1
    yield _check_assert_deriv_0d_exp, 10, (10, ), 10
    yield _check_assert_deriv_0d_exp, 10, (3, 4), 12
    yield _check_assert_deriv_0d_exp, 10, (2, 3, 4), 24


def _check_assert_deriv_nd(nrep, arg_shape, output_mask, numtested):
    _function = lambda arg: 0.5 * arg**2

    def _gradient(arg):
        result = np.zeros(arg_shape + arg_shape)
        for idx, val in np.lib.index_tricks.ndenumerate(arg):
            result[idx + idx] = val
        return result

    for _ in range(nrep):
        origin = np.random.normal(0, 1, arg_shape)
        assert assert_deriv(_function, _gradient, origin, output_mask=output_mask) == numtested


def test_assert_deriv_nd():
    yield _check_assert_deriv_nd, 10, (), None, 1
    yield _check_assert_deriv_nd, 10, (10, ), None, 100
    yield _check_assert_deriv_nd, 10, (3, 4), None, 144
    yield _check_assert_deriv_nd, 10, (2, 3, 4), None, 576
    yield _check_assert_deriv_nd, 10, (5, ), np.array([1, 1, 0, 0, 1], dtype=bool), 15
    yield _check_assert_deriv_nd, 10, (2, 2), np.array([[1, 0], [0, 1]], dtype=bool), 8


def _check_assert_deriv_extra1(nrep, arg_shape, output_mask, numtested):
    _function = lambda arg: 0.5 * (arg**2).sum(axis=-1)

    def _gradient(arg):
        result = np.zeros(arg_shape[:-1] + arg_shape, float)
        for idx, val in np.lib.index_tricks.ndenumerate(arg):
            result[idx[:-1] + idx] = val
        return result

    for _ in range(nrep):
        arg = np.random.normal(0, 1, arg_shape)
        assert assert_deriv(_function, _gradient, arg, output_mask=output_mask) == numtested


def test_assert_deriv_extra1():
    yield _check_assert_deriv_extra1, 10, (3,), None, 3
    yield _check_assert_deriv_extra1, 10, (4, 3), None, 48
    yield _check_assert_deriv_extra1, 10, (2, 4, 3), None, 192
    yield _check_assert_deriv_extra1, 10, (4, 3), np.array([1, 0, 0, 1], dtype=bool), 24
    yield _check_assert_deriv_extra1, 10, (2, 2, 3), np.array([[1, 1], [0, 0]], dtype=bool), 24


def _check_assert_deriv_nd_zeros(nrep, arg_shape, output_mask, numtested):
    _function = lambda arg: np.ones(arg_shape)
    _gradient = lambda arg: np.zeros(arg_shape + arg_shape)
    for _ in range(nrep):
        args = np.random.normal(0, 1, arg_shape)
        assert assert_deriv(_function, _gradient, args, output_mask=output_mask) == numtested


def test_assert_deriv_nd_zeros():
    yield _check_assert_deriv_nd_zeros, 10, (), None, 1
    yield _check_assert_deriv_nd_zeros, 10, (10, ), None, 100
    yield _check_assert_deriv_nd_zeros, 10, (3, 4), None, 144
    yield _check_assert_deriv_nd_zeros, 10, (2, 3, 4), None, 576
    yield _check_assert_deriv_nd_zeros, 10, (5, ), np.array([0, 1, 1, 0, 0], dtype=bool), 10
    yield _check_assert_deriv_nd_zeros, 10, (2, 2), np.array([[0, 1], [1, 0]], dtype=bool), 8


def test_assert_deriv_nd_weights():

    # function is indeterminate for arg[0] <= 1
    def _function(arg):
        with np.errstate(divide='raise'):
            return arg[1] / max(0, arg[0] - 1) + arg[2]

    # gradient is indeterminate for arg[0] <= 1
    def _gradient(arg):
        with np.errstate(divide='raise'):
            return np.array([-arg[1] / (arg[0] - 1)**2, 1 / max(0, arg[0] - 1), 1.0])

    # do searches near the indeterminate region
    arg = np.array([1.03, 4.0, 1.0])
    # romin searches into arg[0] < 1
    with assert_raises(FloatingPointError):
        assert_deriv(_function, _gradient, arg, 0.1)
    # reduce widths on arg[0] so that romin does not search so far
    widths = np.array([1.e-4, 1.0, 1.0])
    assert assert_deriv(_function, _gradient, arg, widths) == 3
    # zero width on arg[0] so that it gets skipped
    widths = np.array([0.0, 1.0, 1.0])
    assert assert_deriv(_function, _gradient, arg, widths) == 2


def test_assert_deriv_corner_cases():
    _function = lambda arg: np.exp(arg).sum()
    arg = np.ones((3, 3))
    with assert_raises(FloatingPointError):
        assert_deriv(_function, np.exp, arg, 0.1, rtol=0, atol=0)


def test_example():
    example_main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Derivcheck is robust and very sensitive tester for analytic derivatives.
# Copyright (C) 2017 Toon Verstraelen <Toon.Verstraelen@UGent.be>.
#
# This file is part of Derivcheck.
#
# Derivcheck is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Derivcheck is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""Very basic usage example of Derivcheck."""


import numpy as np

from derivcheck import assert_deriv


def function(arg):
    """Compute a trivial quadratic function."""
    return 0.5*(arg**2).sum()


def gradient(arg):
    """Compute analytic partial derivatives of ``function``."""
    return arg


def main():
    """Run the example."""
    # Some reference pointsat which the derivative must be tested.
    origin = np.random.normal(0, 1, (4, 3))

    # Test the derivatives at the reference points. See docstring for optional arguments.
    assert_deriv(function, gradient, origin)


if __name__ == '__main__':  # pragma: no cover
    main()



# -*- coding: utf-8 -*-
# Derivcheck is robust and very sensitive tester for analytic derivatives.
# Copyright (C) 2017 Toon Verstraelen <Toon.Verstraelen@UGent.be>.
#
# This file is part of Derivcheck.
#
# Derivcheck is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Derivcheck is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""Robust and sensitive tester for first-order analytic partial derivatives."""

from __future__ import print_function

import numpy as np

from .version import __version__

__all__ = ['diff_ridders', 'assert_deriv']


def diff_ridders(function, origin, stepsize, con=1.4, safe=2.0, maxiter=15):
    """Estimate first-order derivative with Ridders' finite difference method.

    This implementation is based on the one from the book Numerical Recipes. The code
    is pythonized and no longer using fixed-size arrays. Also, the output of the function
    can be an array.

    Parameters
    ----------
    function : function
        The function to be differentiated.
    origin : float
        The point at which must be differentiated.
    stepsize : float
        The initial step size.
    con : float
        The rate at which the step size is decreased (contracted). Must be larger than
        one.
    safe : float
        The safety check used to terminate the algorithm. If Errors between successive
        orders become larger than ``safe`` times the error on the best estimate, the
        algorithm stop. This happens due to round-off errors.
    maxiter : int
        The maximum number of iterations, equals the maximum number of function calls and
        also the highest polynomial order in the Neville method.

    Returns
    -------
    estimate : float
        The best estimate of the first-order derivative.
    error : float
        The (optimistic) estimate of the error on the derivative.

    """
    if stepsize == 0.0:
        raise ValueError('stepsize must be nonzero.')
    if con <= 1.0:
        raise ValueError('con must be larger than one.')
    if safe <= 1.0:
        raise ValueError('safe must be larger than one.')

    con2 = con*con
    table = [[(
        np.asarray(function(origin + stepsize))
        - np.asarray(function(origin - stepsize))
    )/(2.0*stepsize)]]
    estimate = None
    error = None

    # Loop based on Neville's method.
    # Successive rows in the table will go to smaller stepsizes.
    # Successive columns in the table go to higher orders of extrapolation.
    for i in range(1, maxiter):
        # Reduce step size.
        stepsize /= con
        # First-order approximation at current step size.
        table.append([(
            np.asarray(function(origin + stepsize))
            - np.asarray(function(origin - stepsize))
        )/(2.0*stepsize)])
        # Compute higher-orders
        fac = con2
        for j in range(1, i+1):
            # Compute extrapolations of various orders, requiring no new
            # function evaluations. This is a recursion relation based on
            # Neville's method.
            table[i].append((table[i][j-1]*fac - table[i-1][j-1])/(fac-1.0))
            fac = con2*fac

            # The error strategy is to compare each new extrapolation to one
            # order lower, both at the present stepsize and the previous one:
            current_error = max(abs(table[i][j] - table[i][j-1]).max(),
                                abs(table[i][j] - table[i-1][j-1]).max())

            # If error has decreased, save the improved estimate.
            if error is None or current_error <= error:
                error = current_error
                estimate = table[i][j]

        # If the highest-order estimate is growing larger than the error on the best
        # estimate, the algorithm becomes numerically instable. Time to quit.
        if abs(table[i][i] - table[i-1][i-1]).max() >= safe * error:
            break
        i += 1
    return estimate, error


class OneDimWrapper(object):
    """Construct a function of a one-dimensional argument from an array function."""

    def __init__(self, function, origin, indices):
        """Initialize a OneDimWrapper object.

        Parameters
        ----------
        function : function
            The original function with an array argument.
        origin : np.ndarray
            The origin, corresponds to the one-dimensional argument equal to 0.
        indices : tuple
            The index of the matrix element of the input array to use

        """
        self.function = function
        self.origin = origin
        self.indices = indices

    def __call__(self, arg1):
        """Compute the one-dimensional function."""
        arg = self.origin.copy()
        arg[self.indices] += arg1
        return self.function(arg)


def assert_deriv(function, gradient, origin, widths=1e-4, output_mask=None, rtol=1e-5, atol=1e-8):
    """Test the gradient of a function.

    Parameters
    ----------
    function : function
        The function whose derivatives must be tested, takes one argument, which may be
        a scalar or an array with shape ``shape_in``. It may also return an array with
        shape ``shape_out``.
    gradient : function
        Computes the gradient of the function, to be tested. It takes one argument, same
        type as ``function``. The return value is an array with shape ``shape_out +
        shape_in``.
    origin : np.ndarray
        The point at which the derivatives are computed.
    widths : float or np.ndarray
        The initial (maximal) step size for the finite difference method. Do not take a
        value that is too small. When an array is given, each matrix element of the input
        of the function gets a different step size. When a matrix element is set to zero,
        the derivative towards that element is not tested. The function will not be
        sampled beyond [origin-widths, origin+widths].
    output_mask : np.ndarray or None
        This option is useful when the function returns an array output: it allows the
        caller to select which components of the output need to be tested. When not given,
        all components are tested.
    rtol : float
        The allowed relative error on the derivative.
    atol : float
        The allowed absolute error on the derivative.

    Returns
    -------
    numtested : int
        The number of derivatives tested.

    Raises
    ------
    AssertionError when the error on the derivative is too large.

    """
    # Make sure origin is always an array object.
    origin = np.asarray(origin)

    # Compute the gradient and give it the 1D or 2D shape. The first index is a raveled
    # output index.
    gradient = np.asarray(gradient(origin))
    if output_mask is not None:
        gradient = gradient[output_mask]
    if origin.ndim == 0:
        gradient = gradient.ravel()
    else:
        gradient = gradient.reshape(-1, origin.size)

    # Flat loop ofer all elements of the input array
    numtested = 0
    for iaxis in range(origin.size):
        # Get the corresponding input array indices.
        if origin.ndim == 0:
            indices = ()
        else:
            indices = np.unravel_index(iaxis, origin.shape)

        # Determine the step size
        if isinstance(widths, float):
            stepsize = widths
        else:
            stepsize = widths[indices]

        # If needed, test this component
        if stepsize > 0:
            # Make a function of only the selected input array element.
            wrapper = OneDimWrapper(function, origin, indices)
            # Compute the numerical derivative of this function and an error estimate.
            deriv_approx, deriv_error = diff_ridders(wrapper, 0.0, stepsize)
            # Get the corresponding analytic derivative.
            deriv = gradient[..., iaxis]
            # Make sure the error on the derivative is smaller than the requested
            # thresholds.
            if deriv_error >= atol and deriv_error >= rtol*abs(deriv).max():
                raise FloatingPointError('Inaccurate estimate of the derivative for '
                                         'iaxis={}.'.format(indices))
            # Flatten the array with numerical derivatives.
            if output_mask is None:
                deriv_approx = deriv_approx.ravel()
            else:
                deriv_approx = deriv_approx[output_mask]
            # Compare
            err_msg = 'derivative toward {} x=analytic y=numeric stepsize={:g}'.format(
                indices, stepsize)
            np.testing.assert_allclose(deriv, deriv_approx, rtol, atol, err_msg=err_msg)
            numtested += deriv.size
    return numtested



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Derivcheck is robust and very sensitive tester for analytic derivatives.
# Copyright (C) 2017 Toon Verstraelen <Toon.Verstraelen@UGent.be>.
#
# This file is part of Derivcheck.
#
# Derivcheck is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Derivcheck is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""Derivcheck setup script.

If you are not familiar with setup.py, just use pip instead:

    pip install derivcheck --user --upgrade

Alternatively, you can install from source with

    ./setup.py install --user
"""

from __future__ import print_function

from setuptools import setup

from tools.gitversion import get_gitversion


def readme():
    """Load README.rst for display on PyPI."""
    with open('README.rst') as f:
        return f.read()


setup(
    name='derivcheck',
    version=get_gitversion('derivcheck', verbose=(__name__ == '__main__')),
    description='A robust and very sensitive tester for analytic derivatives.',
    long_description=readme(),
    author='Toon Verstraelen',
    author_email='Toon.Verstraelen@UGent.be',
    url='https://github.com/theochem/derivcheck',
    packages=['derivcheck'],
    install_requires=['numpy', 'nose'],
    python_requires='>=2.7',
    zip_safe=False,
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ],
)
